# -*- coding: utf-8 -*-
############################################################################
#
# Copyright © 2014, 2015, 2016 OnlineGroups.net and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
# Original software part of https://github.com/groupserver/gs.dmarc/
#
# This has been modified from the original software.
# 01/24/2017 Valimail Inc
# Contact: Gene Shuman <gene@valimail.com>
#
############################################################################
from __future__ import absolute_import, unicode_literals, print_function
import importlib.resources
try:
    # typing is needed by mypy, but is unused otherwise
    from typing import Dict, Text  # noqa: F401
except ImportError:
    pass
from dns.resolver import (resolve, NXDOMAIN, NoAnswer, NoNameservers)
try:
    from publicsuffix2 import PublicSuffixList
except ImportError:
    # Fall back to deprecated publicsuffix if publicsuffix2 is not available
    from publicsuffix import PublicSuffixList
import sys
from collections import OrderedDict

class DMARCException(Exception):
    """Base class for DMARC errors."""
    pass

def answer_to_dict(answer):
    # type: (Text) -> Dict[unicode, unicode]
    '''Turn the DNS DMARC answer into a dict of tag:value pairs.
    Examples:
    >>> answer_to_dict("v=DMARC1; p=reject; rua=mailto:dmarc.reports@valimail.com,mailto:dmarc_agg@vali.email; ruf=mailto:dmarc.reports@valimail.com,mailto:dmarc_c0cb7153_afrf@vali.email")
    {'v': 'DMARC1', 'p': 'reject', 'rua': 'mailto:dmarc.reports@valimail.com,mailto:dmarc_agg@vali.email', 'ruf': 'mailto:dmarc.reports@valimail.com,mailto:dmarc_c0cb7153_afrf@vali.email'}
    >>> answer_to_dict("v=DMARC1; p=none; sp=reject")
    {'v': 'DMARC1', 'p': 'none', 'sp': 'reject'}
    >>> answer_to_dict('"v=DMARC1; p=none; rua=mailto:dmarc.wheaton.rua@wheaton.edu,mailto:dmarc_agg@waybettermarketing.com; ruf=mailto:dmarc.wheaton.ruf@wheaton.edu,mailto:dmarc_fr@waybettermarketing.com; fo=1;" ""')
    {'v': 'DMARC1', 'p': 'none', 'rua': 'mailto:dmarc.wheaton.rua@wheaton.edu,mailto:dmarc_agg@waybettermarketing.com', 'ruf': 'mailto:dmarc.wheaton.ruf@wheaton.edu,mailto:dmarc_fr@waybettermarketing.com', 'fo': '1'}
    >>> try:
    ...     answer_to_dict("v=DMARC1;p=none;pct=100;rua=mailto:postmaster@somenet.org;mailto:postmaster@somenet.org;ri=3600;fo=1;")
    ... except DMARCException as a:
    ...     print('dmarc_lookup.DMARCException:', a)
    dmarc_lookup.DMARCException: missing tag or value: v=DMARC1;p=none;pct=100;rua=mailto:postmaster@somenet.org;mailto:postmaster@somenet.org;ri=3600;fo=1;
    '''

    a = answer.strip('" ')
    rawTags = [t.split('=') for t in a.split(';') if t]
    try:
        retval = {t[0].strip().lower(): t[1].strip().lower() for t in rawTags}
    except IndexError:
        raise DMARCException('missing tag or value: {0}'.format(answer))
    # Simpler to lowercase everything and put 'v' back.  Already validated
    # before answer_to_dict is called, so should be fine.
    retval['v'] = 'DMARC1'
    return retval

def dns_query(name, qtype='TXT'):
    try:
        return resolve(name, qtype)
    except (NXDOMAIN, NoAnswer, NoNameservers):
        return None

def lookup_receiver_record(host, dnsfunc=dns_query):
    # type: (str), dnsfunc(optional) -> Dict[unicode, unicode]
    '''Lookup the reciever policy for a host
    :param str host: The host to query. The *actual* host that is queried has
                 ``_dmarc.`` prepended to it.
    :param dnsfunc.  a function from domain names to txt records for DNS lookup
    :returns: The DMARC receiver record for the host. If there is no published
          record then None is returned.
    :rtype: {tag => value}
    '''

    dmarcHost = '_dmarc.{0}'.format(host)

    answer = dnsfunc(dmarcHost)

    # This is because dns_query returns a dns.resolver.Answer object while the
    # test suite dnsfunc returns a string (which does not have quotes on it
    # like the dns.resolver object).
    if dnsfunc != dns_query:
        if answer:
            answer = ['"' + answer + '"']

    if not answer:
        return {}
    else:
        # One might think the only TXT record at _dmarc would be a DMARC
        # record, but one would be wrong.  We need to check all the answers,
        # not just the first.  Also, we can addrss RFC 7489, Section 6.6.3,
        # Policy Discovery reuls to treat two DMARC records like none.
        tags = False
        for result in answer:
            # Check that v= field is the first one in the answer (which is in
            # double quotes) as per Section 7.1 (5):
            #     In particular, the "v=DMARC1" tag is mandatory and MUST appear
            #     first in the list. Discard any that do not pass this test.
            # http://tools.ietf.org/html/draft-kucherawy-dmarc-base-04#section-7.1
            if str(result)[:9] == '"v=DMARC1':
                if not tags:
                    tags = answer_to_dict(str(result))
                else:
                    return {} # Multiple DMARC records
        if tags:
            return tags # One DMARC record worth of tags
        else:
            return {} # No DMARC records


def receiver_record(host, dnsfunc=dns_query):
    # type: (str), dnsfunc(optional) -> (Dict[unicode, unicode], is_subdomain)
    '''Get the DMARC receiver record for a host.
    :param str host: The host to lookup.
    :param dnsfunc.  a function from domain names to txt records for DNS lookup
    :returns: The DMARC reciever record for the host.
    :rtype:  A dict of {tag => value} results

    The :func:`receiver_record` function looks up the DMARC reciever record
    for ``host``. If the host does not have a pubished record
    `the organizational domain`_ is determined. The DMARC record for the
    organizational domain is queried
    (if specified) or the overall record for the domain is returned.
    '''
    hostSansDmarc = host if host[:7] != '_dmarc.' else host[7:]

    retval = lookup_receiver_record(hostSansDmarc, dnsfunc)
    if retval:
        return (retval, False)

    # lookup for org_domain
    newHost = get_org_domain(host)
    retval = lookup_receiver_record(newHost, dnsfunc)

    return (retval, newHost)

def receiver_record_walk(host, dnsfunc=dns_query):
    # type: (str), dnsfunc(optional) -> (Dict[unicode, unicode])
    '''Get the DMARC receiver record for a host using the DMARCbis-07 tree
    walk.
    :param str host: The host to lookup.
    :param dnsfunc.  a function from domain names to txt records for DNS lookup
    :returns: The DMARC reciever record for the host.
    :rtype:  A dict of {tag => value} results

    The :func:`receiver_record_walk` function looks up the DMARC reciever
    record for ``host``. If the host does not have a pubished record, the DNS
    tree is walked up until a record is found or the tree is exhausted.
    Specific types of lookups such as organizational domain or PSD are no
    longer relevant.

    If multiple records are returned, the first (longest match) is the policy
    record.  The last, non-PSD (no psd=y flag) is the organizational domain
    for alignment determination.

    Return a list of results for each step in the tree walk.
    '''
    hostSansDmarc = host if host[:7] != '_dmarc.' else host[7:]

    result = OrderedDict()
    retval = lookup_receiver_record(hostSansDmarc, dnsfunc)
    if retval:
        result[hostSansDmarc] = retval

    # walk the tree
    tree = hostSansDmarc.split('.')
    if len(tree) < 5:
        level = len(tree) - 1
    else:
        level = 4
    while level > 0:
        newHost = '.'.join(tree[(len(tree) - level):len(tree)])
        level -= 1
        retval = lookup_receiver_record(newHost, dnsfunc)
        if retval:
            result[newHost] = retval
    return result


def get_org_domain_from_suffix_list(location, domain):
    with open(location) as suffixList:
        psl = PublicSuffixList(suffixList)
        return psl.get_public_suffix(domain)


def get_org_domain(domain):
    try:
        from authheaders.findpsl import location
        return get_org_domain_from_suffix_list(location, domain)
    except ModuleNotFoundError:
        ref = importlib.resources.files('authheaders') / 'public_suffix_list.txt'
        with importlib.resources.as_file(ref) as location:
            return get_org_domain_from_suffix_list(location, domain)

def _test():
    import doctest, dmarc_lookup
    return doctest.testmod(dmarc_lookup)


if __name__ == '__main__':
    _test()



#!/usr/bin/python3


import argparse
import sys
import authheaders


def main():
    """ Find DMARC policy for a domain.

    There are up to three different conventions for DMARC related policy
    discovery:
        1.  RFC 7489: domain + org domain
        2.  RFC 9091: domain + ord domain + PSD domain
        3.  dmarcbis-04 [1]: domain + walk

    This script checks to see which policy is relevant to a domain based on
    these methods.
    @param domain: Policy domain to locate.
    @param s: Policy selection logic to use, default DMARC.
    @param v: Verbose output:
      [original_from, policydomain, result_comment, policy, record, orgdomain]
    @param q: Quiet output

    [1] https://datatracker.ietf.org/doc/draft-ietf-dmarc-dmarcbis Section 4.5
        DNS Tree Walk
    """
    parser = argparse.ArgumentParser(
        description='Find DMARC policy for a domain.',)
    parser.add_argument('domain', action="store",
                        help='Usually From: domain of an email')
    parser.add_argument('-s', '--select',
                        choices=['DMARC', 'PSD', 'DMARCbis'], default='DMARC',
                        help='Select policy discovery method: Default is \
                        DMARC')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='More details: [original_from, policydomain, \
                        result_comment, policy, record, orgdomain]')
    parser.add_argument('-q', '--quiet', action='store_true', default=False,
                        help='turn quiet mode on.  Exit 0 == policy found.  \
                        Exit 1 == no policy found.')
    args=parser.parse_args()

    if args.select == 'DMARC':
        psddmarc = False
        dmarcbis = False
    elif args.select == 'PSD':
        psddmarc=True
        dmarcbis=False
    elif args.select == 'DMARCbis':
        psddmarc=False
        dmarcbis=True

    res = authheaders.dmarc_per_from(args.domain, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=psddmarc, dmarcbis=dmarcbis, policy_only=True)
    if not args.quiet:
        if args.verbose:
            print(res)
        else:
            print('From Domain: {0} Policy Domain: {1}, Policy: {2} Organizational Domain: {3}'.format(res[0], res[1], res[3], res[5]))
    else:
        if res[1] == 'None':
            sys.exit(1)
        else:
            sys.exit(0)

if __name__ == '__main__':
  main()




# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the author be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#
# Copyright (c) 2017 Valimail Inc
# Contact: Gene Shuman <gene@valimail.com>
#

import importlib.resources
import re
import sys
import copy
from email.utils import getaddresses
from authheaders.dmarc_lookup import dns_query, receiver_record, receiver_record_walk, get_org_domain
from authres import SPFAuthenticationResult, DKIMAuthenticationResult, AuthenticationResultsHeader
from authres.arc import ARCAuthenticationResult
from authres.dmarc import DMARCAuthenticationResult
from dkim import ARC, DKIM, arc_verify, dkim_verify, DKIMException, rfc822_parse
from dns.exception import DNSException

# Please accept my appologies for doing this
try:
    import spf
except ImportError:
    pass

__all__ = [
    "authenticate_message",
    "sign_message",
    "chain_validation"
    ]


def get_domain_part(address):
    '''Return domain part of an email address'''
    if sys.version_info < (3, 0) and isinstance(address, str):
        address = bytes(address)
    elif isinstance(address, str):
        address = bytes(address, 'utf8')
    res = re.findall(b'@([a-z0-9.]+)', address, re.IGNORECASE)
    return res[0].lower().decode('ascii')


def check_psddmarc_list(psdname, dnsfunc=dns_query):
    """Check psddmarc.org list of PSD DMARC participants"""
    try:
        # If the PSD registry is locally available, use it.
        ref = importlib.resources.files('authheaders') / 'psddmarc.csv'
        with importlib.resources.as_file(ref) as psdfile_name:
            with open(psdfile_name) as psd_file:
                psds = []
                for line in psd_file.readlines():
                    sp = line.split(',')
                    if sp[1] == 'current':
                        psds.append(sp[0][1:])
                if psdname in psds:
                    return True
                else:
                    return False
    except:
        # If not, use the DNS query list.
        psd_list_host = '.psddmarc.org'
        psd_lookup = psdname + psd_list_host
        answer = dnsfunc(psd_lookup)
        if answer:
            return True
        else:
            return False


def dmarc_per_from(from_domain, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False, dmarcbis=False, policy_only=False):
    """DMARC result for a single From domain."""
    original_from = from_domain
    psddomain = False
    if not dmarcbis: # It's all different in the future
        # Get dmarc record for domain
        if(dnsfunc):
            record, orgdomain = receiver_record(from_domain, dnsfunc=dnsfunc)
        else:
            record, orgdomain = receiver_record(from_domain)
        # Report if DMARC record is From Domain or Org Domain
        if record and orgdomain:
            result_comment = 'Used Org Domain Record'
            policydomain = orgdomain
        elif record:
            result_comment = 'Used From Domain Record'
            orgdomain = from_domain
            policydomain = None

        # Get psddmarc record if doing PSD DMARC, no DMARC record, and PSD is
        #  listed
        if (not record) and psddmarc:
            org_domain = get_org_domain(from_domain)
            if(dnsfunc):
                if check_psddmarc_list(org_domain.split('.',1)[-1],
                                       dnsfunc=dnsfunc):
                    record, _ = receiver_record(org_domain.split('.',1)[-1],
                                                dnsfunc=dnsfunc)
            else:
                if check_psddmarc_list(org_domain.split('.',1)[-1]):
                    record, _ = receiver_record(org_domain.split('.',1)[-1])
            if record:
                psddomain = org_domain.split('.',1)[-1]
                result_comment = 'Used Public Suffix Domain Record'
    else:
        # Get dmarc record for domain (tree walk)
        # TODO: Not very efficient.  Always does tree walk, even if not
        # really needed.  receiver_record_walk does sequential DNS lookups
        # vice parallel.
        orgdomain = False
        record = None
        if(dnsfunc):
            treeresults = receiver_record_walk(from_domain, dnsfunc=dnsfunc)
        else:
            treeresults = receiver_record_walk(from_domain)
        # Then find org domain, per DMARCbis 07
        # Fake psd=yes (since no one publishes this yet) - FIXME later
        for dmn, rec in reversed(list(treeresults.items())):
            if check_psddmarc_list(dmn):
                treeresults[dmn]['psd'] = 'y'
                break
        orgresults = copy.deepcopy(treeresults)
        psddomain = False
        for dmn, rec in list(treeresults.items()):
            try:
                psd = treeresults[dmn]['psd']
                # For psd=y, use next longer domain as org.
                if psd == 'y':
                    if dmn == from_domain:
                        orgdomain = dmn
                        record = rec
                        result_comment = 'Used From Domain which is also PSD'
                    else:
                        psddots = len(dmn[0].split('.'))
                        fromsplit = from_domain.split('.')
                        orgdomain = ''
                        start = True
                        for part in fromsplit[(len(fromsplit) - psddots - 1):]:
                            if not start:
                                orgdomain += '.'
                            start = False
                            orgdomain += part
                        try:
                            if treeresults[orgdomain]:
                                record = treeresults[orgdomain]
                        except:
                            record = rec
                            psddomain = dmn
                        result_comment = 'Used Tree Walk, org one level below PSD'
                    break
                # For psd=n, this is the org domain.
                if psd == 'n':
                    orgdomain = dmn
                    record = rec
                    result_comment = 'Used Tree Walk Record which is PSD=n'
                    break
            except:
                # No PSD tag
                pass
        if not orgdomain:
            # If psd=u (implicit or explicit), usual case.
            for dmn, rec in reversed(list(treeresults.items())):
                if rec and dmn != from_domain:
                    orgdomain = dmn
                    record = rec
                    result_comment = 'Used Tree Walk Record'
                    break
        if not orgdomain:
            # If all else fails, the From domain is the org domain.
            orgdomain = from_domain
            try:
                record = treeresults[from_domain]
                result_comment = 'Used From Domain Record'
            except:
                record = False
                result_comment = 'From domain has no DMARC record'


    if record and record.get('p'): # DMARC P tag is mandatory
        # find policy
        policy = record['p']
        if policy[-1:] == '\\':
            policy = policy[:-1]
        try:
            sp = record['sp']
            if sp[-1:] == '\\':
                sp = sp[:-1]
        except KeyError:
            sp = policy
        try:
            np = record['np']
            if np[-1:] == '\\':
                np = np[:-1]
        except KeyError:
            np = None

        if orgdomain or psddomain:
            if np:
                exists = False
                for qtype in ['a', 'mx', 'aaaa']:
                    if(dnsfunc):
                        res = dnsfunc(from_domain)
                    else:
                        res = dns_query(from_domain, qtype)
                    if res:
                        exists = True
                        break
                if exists:
                    policy = sp
                else:
                    policy = np
            else:
                policy = sp

        adkim = record.get('adkim', 'r')
        aspf  = record.get('aspf',  'r')

        # get result
        result = "fail"
        if not policy_only and spf_result and spf_result.result == "pass":
            # The domain in SPF results often includes the local part, even though
            # generally it SHOULD NOT (RFC 7601, Section 2.7.2, last paragraph).
            try:
                mail_from_domain = get_domain_part(spf_result.smtp_mailfrom)
            except IndexError:
                mail_from_domain = None
            spf_result.smtp_mailfrom = mail_from_domain
            if aspf == "s" and from_domain == mail_from_domain:
                result = "pass"
            elif aspf == "r" and get_org_domain(from_domain) == get_org_domain(mail_from_domain):
                result = "pass"

        if not policy_only and dkim_result and dkim_result.result == "pass":
            if adkim == "s" and from_domain == dkim_result.header_d:
                result = "pass"
            elif adkim == "r" and get_org_domain(from_domain) == get_org_domain(dkim_result.header_d):
                result = "pass"
    else:
        # If no DMARC record, no result
        result = 'none'
        result_comment = ''
        from_domain = ''
        policy = ''

    if policy_only:
        if psddomain:
            policydomain = psddomain
        else:
            policydomain = orgdomain
        if not record:
            result_comment = "None"
            policydomain = None
        results = [original_from, policydomain, result_comment, policy, record, orgdomain]
        return(results)
    else:
        return(result, result_comment, from_domain, policy)

def check_spf(ip, mail_from, helo):
    res, reason = spf.check2(ip, mail_from, helo)
    if res is not None:
        return SPFAuthenticationResult(result=res, reason=reason, smtp_mailfrom=mail_from, smtp_helo=helo)
    else:
        return SPFAuthenticationResult(result=None, reason=None, smtp_mailfrom=mail_from, smtp_helo=helo)


def check_dkim(msg, dnsfunc=None):
    try:
        d = DKIM(msg)
        if(dnsfunc):
            res = d.verify(dnsfunc=dnsfunc) and 'pass' or 'fail'
        else:
            res = d.verify() and 'pass' or 'fail'
    except DKIMException as e:
        res = 'fail'
    except DNSException as e:
        res = 'temperror'
    except Exception as e:
        res = 'fail'

    header_i = d.signature_fields.get(b'i', b'').decode('ascii')
    header_d = d.signature_fields.get(b'd', b'').decode('ascii')
    if res:
        return DKIMAuthenticationResult(result=res, header_d=header_d, header_i=header_i)
    else:
        return DKIMAuthenticationResult(result=None)


def check_arc(msg, logger=None, dnsfunc=None):
    """ Compute the chain validation status of an inbound message.
    @param msg: an RFC822 formatted message (with either \\n or \\r\\n line endings)
    @param logger: An optional logger
    @param dnsfunc: An optional dns lookup function (intended for testing)
    """

    a = ARC(msg)
    try:
        if(dnsfunc):
            cv, results, comment = a.verify(dnsfunc=dnsfunc)
        else:
            cv, results, comment = a.verify()
    except DKIMException as e:
        cv, results, comment = CV_Fail, [], "%s" % e
    except DNSException as e:
        cv, results, comment = CV_Fail, [], "%s" % e
    except Exception as e:
        cv, results, comment = CV_Fail, [], "%s" % e
    if comment == 'success':
        comment = None
    if cv is not None:
        return ARCAuthenticationResult(result=cv.decode('ascii'), result_comment=comment)
    else:
        return ARCAuthenticationResult(result='none', result_comment=comment)


def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False, dmarcbis=False):

    # get from domain
    headers, _ = rfc822_parse(msg)
    from_headers = [a[1] for a in getaddresses(x[1].replace(b'\r\n', b'')
        .decode(errors='ignore').strip() for x in headers
        if x[0].lower() == b"from")]

    if len(from_headers) > 1:
        # multi-from processing per RFC 7489 6.6.1
        domain_results = []
        for from_header in from_headers:
            try:
                from_domain = get_domain_part(from_header)
            except IndexError:
                result = 'permerror'
                result_comment = 'Unable to extract From domain: {0}'.format(from_header)
                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from='none', policy='none')
            try:
                domain_results.append(dmarc_per_from(from_domain, spf_result=spf_result, dkim_result=dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc, dmarcbis=dmarcbis))
            except dmarc_lookup.DMARCException as result_comment:
                result = 'permerror'
                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy='none')

        for domain in domain_results:
            if domain[3] == 'reject':
                result, result_comment, from_domain, policy = domain
                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy=policy)
        for domain in domain_results:
            if domain[3] == 'quarantine':
                result, result_comment, from_domain, policy = domain
                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy=policy)
        for domain in domain_results:
            if domain[3] == 'none':
                result, result_comment, from_domain, policy = domain
                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy=policy)
        result, result_comment, from_domain, policy = domain
    elif len(from_headers) == 1:
        from_header =  from_headers[0]
        try:
            from_domain = get_domain_part(from_header)
        except IndexError:
            result = 'permerror'
            result_comment = 'Unable to extract From domain: {0}'.format(from_header)
            return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from='none', policy='none')
        try:
            result, result_comment, from_domain, policy = dmarc_per_from(from_domain, spf_result=spf_result, dkim_result=dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc, dmarcbis=dmarcbis)
        except dmarc_lookup.DMARCException as result_comment:
            result = 'permerror'
            return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy='none')

    else:
        result = 'none'

    if result != 'none':
        return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy=policy)
    else:
        return DMARCAuthenticationResult(result=result, header_from=from_domain)


def authenticate_message(msg, authserv_id, prev=None, spf=False, dkim=True, arc=False, dmarc=True, ip=None, mail_from=None, helo=None, dnsfunc=None, psddmarc=False, dmarcbis=False):
    """Authenticate an RFC822 message and return the Authentication-Results header
    @param msg: an RFC822 formatted message (with either \\n or \\r\\n line endings)
    @param authserv_id: The id of the server performing the authentication
    @param prev: an existing authentication results header to append results to
    @param spf: Perform SPF check
    @param dkim: Perform DKIM check
    @param dmarc: Perform DMARC check
    @param psddmarc: Perform PSD DMARC check (RFC 9091)
    @param dmarcbis: Use DMARCbis policy discovery and alignment
    @param arc: Perform ARC chain validation check
    @param ip: (SPF) IP address of incoming request
    @param mail_from: (SPF) Sender declared in MAIL FROM
    @param helo: (SPF) EHLO/HELO domain of incoming message
    @param dnsfunc: An optional dns lookup function (intended for testing)
    @return: The Authentication-Results header
    """

    if spf and 'spf' not in sys.modules:
        raise Exception('pyspf must be installed manually for spf authentication')

    results = []
    if prev:
        arobj = AuthenticationResultsHeader.parse(prev)
        results = arobj.results

    spf_result  = next((x for x in results if type(x) == SPFAuthenticationResult), None)
    dkim_result = next((x for x in results if type(x) == DKIMAuthenticationResult), None)
    arc_result  = next((x for x in results if type(x) == ARCAuthenticationResult), None)

    if spf and not spf_result:
        spf_result = check_spf(ip, mail_from, helo)
        results.append(spf_result)

    if dkim and not dkim_result:
        dkim_result = check_dkim(msg, dnsfunc=dnsfunc)
        results.append(dkim_result)

    if arc and not arc_result:
        arc_result = check_arc(msg, None, dnsfunc=dnsfunc)
        results.append(arc_result)

    if dmarc:
        dmarc_result = check_dmarc(msg, spf_result, dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc, dmarcbis=dmarcbis)
        results.append(dmarc_result)

    auth_res = AuthenticationResultsHeader(authserv_id=authserv_id, results=results)
    return str(auth_res)

def sign_message(msg, selector, domain, privkey, sig_headers, sig='DKIM', srv_id=None,
                 identity=None, length=None, canonicalize=(b'relaxed', b'relaxed'), timestamp=None,
                 logger=None, standardize=False):
    """Sign an RFC822 message and return the ARC or DKIM header(s)
    @param msg: an RFC822 formatted message (with either \\n or \\r\\n line endings)
    @param selector: the DKIM selector value for the signature
    @param domain: the DKIM domain value for the signature
    @param privkey: a PKCS#1 private key in base64-encoded text form
    @param sig_headers: a list of strings indicating which headers are to be signed
    @param sig: "DKIM" or "ARC"
    @param srv_id: an authserv_id to identify AR headers to sign
    @param identity: (DKIM) the DKIM identity value for the signature (default "@"+domain)
    @param length: (DKIM) true if the l= tag should be included to indicate body length (default False)
    @param canonicalize: (DKIM) the canonicalization algorithms to use (default (Relaxed, Relaxed))
    @param timestamp: (for ARC testing) a manual timestamp to use for ARC signature generation
    @param logger: An optional logger
    @param standardize: A testing flag for arc to output a standardized header format
    @return: The DKIM-Message-Signature, or ARC set headers
    @raises: DKIMException if mis-configured
    """

    if sig=="DKIM":
        return DKIM(msg, logger=logger).sign(selector, domain, privkey, include_headers=sig_headers,
                              identity=identity, length=length, canonicalize=canonicalize)
    else:
        return ARC(msg, logger=logger).sign(selector, domain, privkey, srv_id, include_headers=sig_headers, timestamp=timestamp, standardize=standardize)
